import os
import csv as csvmod
import time
import click
import random
from talc.evals import EvalsClient, TestCaseResponse, TestCaseWithRubric


talc_api_key: str | None = os.getenv("TALC_API_KEY", None)
talc_base_url: str | None = os.getenv("TALC_BASE_URL", "http://localhost:8000")


@click.group()
def evals():
    pass


@evals.command()
@click.option("--count", default=0, help="Number of cases to sample (0 for all).")
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option("--dataset", prompt="Dataset ID", help="The ID of the dataset to load.")
@click.option("--csv", default=None, help="Path to CSV file to save results to.")
def getdataset(count: int, dataset: str, api_key: str, csv: str) -> None:
    """Download or pretty print a test dataset from the server."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url)
    data = client.get_dataset(dataset)

    if count == 0:
        count = len(dataset)

    random.shuffle(data)

    if csv is not None:
        with open(csv, "w", newline="") as csvfile:
            writer = csvmod.writer(csvfile)
            writer.writerow(["id", "content"])
            for case in data[:count]:
                writer.writerow([case.id, case.question])

    for case in data[:count]:
        print(case.question)


@evals.command()
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option(
    "--csv",
    prompt="Results CSV path",
    help="Path to CSV file to load results from. Columns should be 'id', 'result'. Header row is optional.",
)
@click.option("--print_failed", is_flag=True, help="Print the failed test cases.")
def grade(api_key: str, csv: str, print_failed: bool):
    """Grade a test run from a csv. Columns should be 'id', 'result'. Header row is optional."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url)

    with open(csv, newline="") as csvfile:
        reader = csvmod.reader(csvfile)

        # Skip the first row. If it's not a header, we add it back later.
        row = next(reader, None)
        if row is None:
            raise click.UsageError("CSV file is empty.")

        results = [TestCaseResponse(id=row[0], response=row[1]) for row in reader]

        # If the first row is not a header, add it back to the results.
        if row[0] != "id" or row[1] != "result":
            results.append(TestCaseResponse(id=row[0], response=row[1]))

    run_info = client.start_run()
    client.submit_responses(results)
    print("Submitted results. Beginning grading for run ID: " + run_info.id)

    while True:
        run_info = client.get_results()
        print(f"Progress: {run_info.completion_progress * 100}%")
        if run_info.completion_progress >= 1:
            print("\n=============================================")
            print("Grading complete.")
            print(f"Accuracy: {run_info.grade_accuracy}")
            print(f"Hallucination: {run_info.grade_hallucination}")
            print("=============================================\n")
            if print_failed and run_info.failed_cases is not None:
                print("Failed cases:")
                for case in run_info.failed_cases:
                    print(
                        f"{case.id}: {case.question} -> {case.response} ({case.grade_reason})"
                    )
            break
        time.sleep(2)


@evals.command()
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option("--run_id", prompt="Run ID", help="The ID of the run to get results for.")
@click.option("--print_failed", is_flag=True, help="Print the failed test cases.")
def getresults(api_key: str, run_id: str, print_failed: bool):
    """Get the grades for a test run."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url, run_id)
    run_grade = client.get_results()

    print(f"Accuracy: {run_grade.grade_accuracy}")
    print(f"Hallucination: {run_grade.grade_hallucination}")

    if print_failed and run_grade.failed_cases is not None:
        print("Failed cases:")
        for case in run_grade.failed_cases:
            print(
                f"{case.id}: {case.question} -> {case.response} ({case.grade_reason})"
            )


@evals.command()
@click.option("--api_key", default=talc_api_key, help="Talc API key")
@click.option(
    "--csv",
    prompt="CSV path",
    help="Path to CSV file to upload. Columns should be 'question' and 'expected answer'.",
)
@click.option("--name", prompt="Dataset name", help="The name of the dataset.")
def adddataset(api_key: str, csv: str, name: str):
    """Add a new dataset to the server."""

    if api_key is None:
        raise click.UsageError(
            "No API key provided. Please set the TALC_API_KEY environment variable or provide the --api_key option."
        )

    client = EvalsClient(api_key, talc_base_url)

    with open(csv, newline="") as csvfile:
        reader = csvmod.reader(csvfile)

        # Skip the first row. If it's not a header, we add it back later.
        header = next(reader, None)
        if header is None:
            raise click.UsageError("CSV file is empty.")

        dataset = [
            TestCaseWithRubric(
                question=row[0], expected_response=row[1], scenario_data={}, id=None
            )
            for row in reader
        ]

        # If the first row is not a header, add it back to the results.
        if header[0] != "question" or header[1] != "expected answer":
            dataset.append(
                TestCaseWithRubric(
                    question=header[0],
                    expected_response=header[1],
                    scenario_data={},
                    id=None,
                )
            )

    dataset_id = client.create_dataset(name, dataset)
    print(f"Dataset created with ID: {dataset_id}")


def main():
    evals()



import json
import logging
import openai
import os
from datetime import datetime
from supabase.client import create_client, Client
from supabase.lib.client_options import ClientOptions
import logging


class _SupabaseClient:
    def __init__(self):
        url: str = os.environ.get(
            "INSTANCE_URL", default="https://qdgodxkfxzzmzwfliahh.supabase.co"
        )
        key: str = os.environ.get(
            "INSTANCE_ANON_KEY",
            default="eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6InFkZ29keGtmeHp6bXp3ZmxpYWhoIiwicm9sZSI6ImFub24iLCJpYXQiOjE2ODg0OTc0NzcsImV4cCI6MjAwNDA3MzQ3N30.4bgCdg77wwOJ9w1hOtCD-z0gBVGv8X_kIxBCr5KDCuA",
        )

        # API key format is organization:api_key
        talc_api_key: str = os.environ.get("TALC_API_KEY", default="")

        if talc_api_key == "":
            logging.warning(
                "TALC_API_KEY environment variable not set. Logging disabled."
            )
            self.__initialized = False
            return

        organization, api_key = talc_api_key.split(":")
        self.__organization: str = organization

        options = ClientOptions(headers={"talckey": api_key})
        self.supabase: Client = create_client(url, key, options=options)
        self.__initialized = True

    def createSession(self):
        if not self.__initialized:
            return None
        response = (
            self.supabase.table("sessions")
            .insert(
                {
                    "organization": self.__organization,
                }
            )
            .execute()
        )
        return response.data[0]["id"]

    def __createInput(
        self, sessionId, generationId, role, content, name, function_call, index
    ):
        response = (
            self.supabase.table("inputs")
            .insert(
                {
                    "session": sessionId,
                    "generation": generationId,
                    "role": role,
                    "content": content,
                    "index": index,
                    "name": name,
                    "function_call": function_call,
                }
            )
            .execute()
        )
        return response.data[0]["id"]

    def __createGeneration(
        self,
        sessionId,
        content,
        function_calls,
        agent,
        generated_at,
        functions_available,
        parameters,
    ):
        response = (
            self.supabase.table("generations")
            .insert(
                {
                    "session": sessionId,
                    "content": content,
                    "functions_called": function_calls,
                    "agent": agent,
                    "generated_at": generated_at,
                    "functions_available": functions_available,
                    "parameters": parameters,
                }
            )
            .execute()
        )
        return response.data[0]["id"]

    def __historyArrayToInputs(self, history, generationId, sessionId):
        for index, chat in enumerate(history):
            encoded_function_call = None
            if "function_call" in chat:
                encoded_function_call = json.dumps(chat["function_call"])

            self.__createInput(
                sessionId,
                generationId,
                chat["role"],
                chat["content"],
                chat.get("name", None),  # Name is optional
                encoded_function_call,
                # Index is reversed because we want the most recent message to have the lowest index
                len(history) - index,
            )

    def log(
        self,
        sessionId,
        history,
        text_content,
        function_calls,
        agent,
        functions_available,
        parameters,
    ):
        generated_at = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")

        generationId = self.__createGeneration(
            sessionId,
            text_content,
            function_calls,
            agent,
            generated_at,
            functions_available,
            parameters,
        )
        self.__historyArrayToInputs(history, generationId, sessionId)


def createSession():
    response = client.createSession()
    return response


def init():
    global client
    client = _SupabaseClient()


globalSession = None


def setGlobalSession(session):
    global globalSession
    globalSession = session


class __alternateCompletion(openai.ChatCompletion):
    @classmethod
    def create(cls, *args, **kwargs):
        # Pop arguments that are not supported by the original create method
        agent = kwargs.pop("agent", "Default")
        session = kwargs.pop("session", globalSession)
        stream = "stream" in kwargs and kwargs["stream"]

        # Handle case where we have received the full response at once.
        if not stream:
            result = super().create(*args, **kwargs)
            try:
                functions_available, parameters = cls.__getFunctionsAndParameters(
                    **kwargs
                )
                text_content, function_calls = cls.__getContent(result.choices)
                if session and agent:
                    client.log(
                        session,
                        kwargs["messages"],
                        text_content,
                        function_calls,
                        agent,
                        functions_available,
                        parameters,
                    )
            except Exception as e:
                logging.warning("Error logging to talc: ", e)

            return result
        # Handle stream case
        else:
            return cls.__streamContent(session=session, agent=agent, *args, **kwargs)

    # Handles the case where we are streaming the response
    # This is really complicated. Godspeed.
    @classmethod
    def __streamContent(cls, *args, **kwargs):
        agent = kwargs.pop("agent", "Default")
        session = kwargs.pop("session", globalSession)

        generator = super().create(*args, **kwargs)

        choices = {}
        final_choice = None

        for result in generator:
            contains_incomplete_option = False
            for index, choice in enumerate(result.choices):
                # Identify if we have at least one incomplete option still streaming
                if choice.finish_reason is None:
                    contains_incomplete_option = True

                if index not in choices:
                    choices[index] = []
                choices[index].append(choice)
            if not contains_incomplete_option:
                final_choice = result
                break  # We have received the full response, break before returning so we can log.
            yield result

        final_content = []
        final_function_calls = []

        # Try not to find bugs in this code, because it's going to be
        # a pain to fix.
        try:
            # Accumulate all the content and function calls for each option
            for option in choices:
                content_accumulator = []
                function_call_accumulator = []
                arguments_accumulator = []
                for chunk in choices[option]:
                    if "delta" in chunk:
                        if "content" in chunk.delta:
                            content_accumulator.append(chunk.delta["content"])
                        if "function_call" in chunk.delta:
                            if "name" in chunk.delta["function_call"]:
                                function_call_accumulator.append(
                                    chunk.delta["function_call"]["name"]
                                )
                            if "arguments" in chunk.delta["function_call"]:
                                arguments_accumulator.append(
                                    chunk.delta["function_call"]["arguments"]
                                )
                if len(content_accumulator) > 0:
                    final_content.append(
                        "".join([x for x in content_accumulator if x != None])
                    )
                else:
                    final_content.append(None)
                if len(function_call_accumulator) > 0:
                    final_function_calls.append(
                        {
                            "name": "".join(
                                [x for x in function_call_accumulator if x != None]
                            ),
                            "arguments": "".join(
                                [x for x in arguments_accumulator if x != None]
                            ),
                        }
                    )
                else:
                    final_function_calls.append(None)
            # Log the accumulated options
            if session and agent:
                functions_available, parameters = cls.__getFunctionsAndParameters(
                    **kwargs
                )
                client.log(
                    session,
                    kwargs["messages"],
                    final_content,
                    final_function_calls,
                    agent,
                    functions_available,
                    parameters,
                )
        except Exception as e:
            logging.warning("Error logging to talc: ", e)

        return final_choice

    @classmethod
    async def __streamContentAsync(cls, *args, **kwargs):
        agent = kwargs.pop("agent", "Default")
        session = kwargs.pop("session", globalSession)

        generator = super().acreate(*args, **kwargs)

        choices = {}
        final_choice = None

        async for result in await generator:
            contains_incomplete_option = False
            for index, choice in enumerate(result.choices):
                # Identify if we have at least one incomplete option still streaming
                if choice.finish_reason is None:
                    contains_incomplete_option = True

                if index not in choices:
                    choices[index] = []
                choices[index].append(choice)
            if not contains_incomplete_option:
                final_choice = result
                break  # We have received the full response, break before returning so we can log.
            yield result

        final_content = []
        final_function_calls = []

        # Try not to find bugs in this code, because it's going to be
        # a pain to fix.
        try:
            # Accumulate all the content and function calls for each option
            for option in choices:
                content_accumulator = []
                function_call_accumulator = []
                arguments_accumulator = []
                for chunk in choices[option]:
                    if "delta" in chunk:
                        if "content" in chunk.delta:
                            content_accumulator.append(chunk.delta["content"])
                        if "function_call" in chunk.delta:
                            if "name" in chunk.delta["function_call"]:
                                function_call_accumulator.append(
                                    chunk.delta["function_call"]["name"]
                                )
                            if "arguments" in chunk.delta["function_call"]:
                                arguments_accumulator.append(
                                    chunk.delta["function_call"]["arguments"]
                                )
                if len(content_accumulator) > 0:
                    final_content.append(
                        "".join([x for x in content_accumulator if x != None])
                    )
                else:
                    final_content.append(None)
                if len(function_call_accumulator) > 0:
                    final_function_calls.append(
                        {
                            "name": "".join(
                                [x for x in function_call_accumulator if x != None]
                            ),
                            "arguments": "".join(
                                [x for x in arguments_accumulator if x != None]
                            ),
                        }
                    )
                else:
                    final_function_calls.append(None)
            # Log the accumulated options
            if session and agent:
                functions_available, parameters = cls.__getFunctionsAndParameters(
                    **kwargs
                )
                client.log(
                    session,
                    kwargs["messages"],
                    final_content,
                    final_function_calls,
                    agent,
                    functions_available,
                    parameters,
                )
        except Exception as e:
            logging.warning("Error logging to talc: ", e)

        yield final_choice
        return

        # print("Content ", option, ":", "".join(content_accumulator))
        # print("Function call ", option, ":", "".join(function_call_accumulator))
        # print("Arguments ", option, ":", "".join(arguments_accumulator))
        # print("=====================================")

    @classmethod
    async def acreate(cls, *args, **kwargs):
        # Pop arguments that are not supported by the original create method
        agent = kwargs.pop("agent", "Default")
        session = kwargs.pop("session", globalSession)
        stream = "stream" in kwargs and kwargs["stream"]

        # Handle case where we have received the full response at once.
        if not stream:
            result = await super().acreate(*args, **kwargs)
            try:
                functions_available, parameters = cls.__getFunctionsAndParameters(
                    **kwargs
                )
                text_content, function_calls = cls.__getContent(result.choices)
                if session and agent:
                    client.log(
                        session,
                        kwargs["messages"],
                        text_content,
                        function_calls,
                        agent,
                        functions_available,
                        parameters,
                    )
            except Exception as e:
                logging.warning("Error logging to talc: ", e)

            return result
        # Handle stream case
        else:
            return cls.__streamContentAsync(
                session=session, agent=agent, *args, **kwargs
            )

    @classmethod
    def __getContent(cls, choices):
        text_content = []
        function_calls = []

        for choice in choices:
            if "function_call" in choice.message:
                function_calls.append(choice.message.function_call)
            else:
                function_calls.append(None)

            if "content" in choice.message:
                text_content.append(choice.message.content)
            else:
                text_content.append(None)

        return text_content, function_calls

    @classmethod
    def __getFunctionsAndParameters(cls, **kwargs):
        functions = None
        if "functions" in kwargs:
            functions = kwargs["functions"]

        ignored_params = ["functions", "messages"]

        parameters = {
            key: val for (key, val) in kwargs.items() if key not in ignored_params
        }

        return json.dumps(functions), json.dumps(parameters)


openai.ChatCompletion = __alternateCompletion



import requests
import json
from pydantic import BaseModel


class TestCase(BaseModel):
    """Represents a single test case that should be run."""

    id: str
    question: str


class TestCaseWithRubric(BaseModel):
    """Represents a single test case that should be run."""

    id: str | None
    question: str
    expected_response: str
    scenario_data: dict[str, object]


class TestCaseResponse(BaseModel):
    """Represents the results of running a single test case."""

    id: str
    response: str


class GradedTestCase(BaseModel):
    """Represents a single test case that has been graded."""

    id: str
    question: str
    response: str
    grade_accuracy: float
    grade_hallucination: float
    grade_reason: str


class RunInfo(BaseModel):
    """Represents a test run."""

    id: str
    completion_progress: float
    grade_accuracy: float | None
    grade_hallucination: float | None
    failed_cases: list[GradedTestCase] | None


class Dataset(BaseModel):
    """Represents a dataset."""

    id: str
    friendly_name: str | None


class EvalsClient:
    """Client for interacting with the Talc evals server."""

    def __init__(self, api_key: str, url: str, run_id: str | None = None) -> None:
        self.url: str = url
        self.run_id: str | None = run_id
        self.api_key: str = api_key

    def start_run(self) -> RunInfo:
        """Create a new test run."""

        response = requests.post(
            f"{self.url}/evals/run/create", headers={"X-TALC-API": self.api_key}
        )
        response.raise_for_status()

        info = RunInfo(**response.json())
        self.run_id = info.id
        return info

    def get_dataset(self, id: str) -> list[TestCase]:
        """Get all test cases in a single dataset from the server."""

        response = requests.get(
            f"{self.url}/evals/dataset/{id}", headers={"X-TALC-API": self.api_key}
        )
        response.raise_for_status()

        return [TestCase(**case) for case in response.json()]

    def submit_responses(self, results: list[TestCaseResponse]) -> None:
        """Submit responses generated by the system for grading."""
        if self.run_id is None:
            raise RuntimeError("No run started. Remember to call start_run() first.")

        data = [x.dict() for x in results if x.id != "id"]

        response = requests.post(
            f"{self.url}/evals/run/{self.run_id}/submit",
            json=data,
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

    def get_results(self) -> RunInfo:
        """Get graded results of a run."""
        if self.run_id is None:
            raise RuntimeError("No run started. Remember to call start_run() first.")

        response = requests.get(
            f"{self.url}/evals/run/{self.run_id}/results",
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

        return RunInfo(**response.json())

    def create_dataset(
        self, friendly_name: str, dataset: list[TestCaseWithRubric]
    ) -> Dataset:
        """Create a new dataset and save it to the database."""

        data = [x.dict() for x in dataset if x.id != "id"]

        response = requests.post(
            f"{self.url}/evals/dataset/create/{friendly_name}",
            json=data,
            headers={"X-TALC-API": self.api_key},
        )
        response.raise_for_status()

        return Dataset(**response.json())
